﻿//*************************************************************************************************
//【サンプルプログラム】『Kinectからデプス画像を取得し、人物部分だけを切り取って表示する』
//【名称】Bitmap_GetPlayerDepthImage　　【Version】1.02
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章
//【作成日】2012年3月5日　【コメント追記】2012年6月30日
//【注意事項】
//　このプログラムをKinect for Windows SDK Version1.5で動かすには、参照設定の
// Microsoft.Kinectを一旦削除し、プロジェクト(P)-参照の追加(F)で、.NETタブ内
// にあるMicrosoft.Kinectを選択・追加してください。 
//*************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;             //【Debug】
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;         //【BitmapData】
using System.Runtime.InteropServices; //【Marshal.Copy】
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択

namespace Bitmap_GetPlayerDepthImage
{
    public partial class Form1 : Form
    {
        //【KINECT関連】
        //　関数内でダイナミックに変数を確保すると動作速度が遅くなるので、ここにまとめて変数を確保しています。
        private KinectSensor KS_KinectSensor;   //【KINECT】
        private DepthImageFrame KS_DepthFrame;  //【KINECTのDepthImageFrame】
        private short[] KS_DepthData;           //【KINECTのDepthImageData】16bit/画素(MSB13bit:Depth, LSB3bit:PlayerID) 
        private byte[] KS_DepthImageDataARGB;   //【KINECTのDepthImageData】32bit/画素,A8R8G8B8表示形式
        private int KS_DepthFrameWidth;         //【KINECTのDepthFrameの横方向画素数】
        private int KS_DepthFrameHeight;        //【KINECTのDepthFrameの縦方向画素数】
        private Bitmap KS_bmp2;                 //【DepthImage表示用Bitmapオブジェクト】
        private BitmapData KS_data2;            //【DepthImage用BitmapData】
        private Rectangle KS_rect2;             //【DepthImage表示枠】
        private int KS_PlayerIndexBitmask;      //【DepthデータのPlayer情報】 0000 0000 0000 0111B = 7
        private int KS_PlayerIndexBitmaskWidth; //【DepthデータのPlayer情報】 3 bits
        private int KS_Lmin;                    //【深度測定可能最近距離】default:  800mm, near: 400mm 
        private int KS_Lmax;                    //【深度測定可能最遠距離】default: 4000mm, near:3000mm
        private bool KS_flagFlip = true;        //【左右反転フラグ】
        private bool KS_flagPlayer = false;     //【Player検出フラグ】
        //【NTL拡張現実センサ関連】---------------------------------------------------------
        private NTL_ColorBar ColorBar = new NTL_ColorBar(); //【デプス値をカラーに変換する部品】
        private string[] BackGroundFileName = { "../../image/MMD.bmp", "../../image/Sky.bmp" };
        
        public Form1()
        {
            InitializeComponent();

            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();
            //【検出結果の表示】
            if (KS_KinectSensor == null) //【Kinect接続が確認できなかった場合】
            {
                pictureBox2.Image = imageList1.Images[1];
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            else //【Kinect検出に成功】
            {
                pictureBox2.Image = imageList1.Images[0];
            }

            //***********************************************************
            //【Kinectの初期化処理】
            //***********************************************************
            //【KINECTの設定】DepthStreamを許可
            KS_KinectSensor.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30); //【DepthStream許可】
            // DepthImageFormatとしては他にResolution320x240Fps30とDepthImageFormat.Resolution80x60Fps30がある
            KS_KinectSensor.SkeletonStream.Enable();                                     //【SkeltonStream許可】
            //【★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★】
            //【★★★デプス情報のPlayerIDが有効になるのはSkeletonStreamを有効にしている場合に限られるので注意★★★】
            //【★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★★】
            //【KINECT動作開始】
            KS_KinectSensor.Start(); //【KINECTスタート】

            //【KINECTデータ取得イベントの設定とデータバッファの確保】********************************************
            KS_KinectSensor.DepthFrameReady += DepthImageReady; //【準備完了でDepthImageReadyがcallされる】
            KS_DepthData = new short[KS_KinectSensor.DepthStream.FramePixelDataLength];              //【深度情報データ用】
            KS_DepthImageDataARGB = new byte[KS_KinectSensor.DepthStream.FramePixelDataLength * 4];  //【深度情報データ表示用】

            KS_PlayerIndexBitmask = DepthImageFrame.PlayerIndexBitmask;
            KS_PlayerIndexBitmaskWidth = DepthImageFrame.PlayerIndexBitmaskWidth;
            KS_DepthFrameWidth = KS_KinectSensor.DepthStream.FrameWidth;   //【DepthImageの横方向画素数】
            KS_DepthFrameHeight = KS_KinectSensor.DepthStream.FrameHeight; //【DepthImageの縦方向画素数】
            KS_bmp2 = new Bitmap(KS_DepthFrameWidth, KS_DepthFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【DepthImage表示用】
            KS_rect2 = new Rectangle(0, 0, KS_DepthFrameWidth, KS_DepthFrameHeight); //【DepthImage用枠】
            KS_KinectSensor.DepthStream.Range = DepthRange.Default;    //【Depth測定モードをDefaultに設定】
            KS_CheckDistance();                                        //【設定結果を表示】Lmax, Lminを取得

            KS_KinectSensor.ElevationAngle = 0; //【KINECTセンサの上下角を0°に設定】設定範囲±27°
            
            ColorBar.CreateColorPalletForAbsolute(KS_Lmax - KS_Lmin + 10); //【カラーパレットを作成する】
            pictureBox3.Image = ColorBar.CreateColorScale(256, 10);    //【256×10画素サイズのカラーバーを作成して表示する】

            pictureBox1.SizeMode = PictureBoxSizeMode.StretchImage; //【キャプチャー画像がpictureBox枠内にはめ込まれます】
            pictureBox1.BackgroundImage = System.Drawing.Image.FromFile(BackGroundFileName[1]); //【Sky.bmpを表示】
        }

        //***************************************************************
        //【CheckDistance】Defaultモード(800mm～4000mm)
        //***************************************************************
        private void radioButton1_CheckedChanged(object sender, EventArgs e)
        {  //【Defaultモード】
            KS_KinectSensor.DepthStream.Range = DepthRange.Default;
            KS_CheckDistance();
        }

        //***************************************************************
        //【距離モード変更】Nearモード(400m～3000mm)
        //***************************************************************
        private void radioButton2_CheckedChanged(object sender, EventArgs e)
        {  //【Nearモード】
            KS_KinectSensor.DepthStream.Range = DepthRange.Near;
            KS_CheckDistance();
        }

        //***************************************************************
        //【CheckDistance】KINECTセンサの検出可能距離現在値を調べる
        //***************************************************************
        private void KS_CheckDistance()
        {
            KS_Lmax = KS_KinectSensor.DepthStream.MaxDepth;
            KS_Lmin = KS_KinectSensor.DepthStream.MinDepth;
            label1.Text = "Min=" + KS_Lmin.ToString() + "mm";
            label2.Text = "Max=" + KS_Lmax.ToString() + "mm";
        }

        //***************************************************************
        //【終了要求ボタン】
        //***************************************************************
        private void button1_Click(object sender, EventArgs e)
        {
            if (KS_KinectSensor != null)
            {
                KS_KinectSensor.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】
            }
            this.Close();              //【Formを閉じる】
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            Int32 angle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if (angle <= KS_KinectSensor.MaxElevationAngle && angle >= KS_KinectSensor.MinElevationAngle)
            {
                KS_KinectSensor.ElevationAngle = angle; //【上下角の変更】
            }
        }

        //***************************************************************
        //【カラー画像の左右反転指定フラグ】
        //***************************************************************
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }

        //***************************************************************
        //【DepthImageの取得イベント】
        //***************************************************************
        private void DepthImageReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (KS_DepthFrame = e.OpenDepthImageFrame())
            {
                if (KS_DepthFrame != null)
                {
                    //kinectSensor.DepthStreamに以下のフラグがある。
                    // .TooNearDepth【距離が近すぎる】, .TooFarDepth【距離が遠すぎる】, .UnknownDepth【距離が不明】
                    KS_DepthFrame.CopyPixelDataTo(KS_DepthData); //【KINECTのDepth情報を配列DepthDataにコピー】
                    KS_ConvertDepthToColor(); //【深度情報をカラーバーの色を使って表示形式をA8R8G8B8形式に変換】

                    KS_data2 = KS_bmp2.LockBits(KS_rect2, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【Lockする】
                    Marshal.Copy(KS_DepthImageDataARGB, 0, KS_data2.Scan0, KS_DepthImageDataARGB.Length); //【UnmanagedMemoryのBitmapDataに配列depthDataをコピーする】
                    KS_bmp2.UnlockBits(KS_data2); //【Lock解除】
                    if (KS_flagFlip) { KS_bmp2.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.Image = KS_bmp2;  //【pictureBox1に画像表示】
                }
                else
                {
                    //【DepthFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【ConvertDepthToColor】16ビットのDepth値をカラーバーに対応する色に変更
        //***************************************************************
        private void KS_ConvertDepthToColor()
        {
            float a;      //【Normalモード】800～4000mm ⇒ 0.0796875  3200色
            int i, j, d;  //【Nearモード】  400～3000mm ⇒ 0.0980769　2600色
            int playerID; //【下位3bitはplayerID】【未使用】
            int f; //【PlayerLost検出用】
            if (KS_KinectSensor.DepthStream.Range == DepthRange.Default) { a = 1.000000f; } else { a = 1.230769f; }
            if (KS_flagPlayer == false) //【player未検出時はDepthImage全体を表示する】
            {
                j = 0;
                for (i = 0; i < KS_DepthData.Length; i++)
                {
                    d = KS_DepthData[i];                  //【距離情報】【上位13bitがDepth値】
                    playerID = d & KS_PlayerIndexBitmask; //【下位3bitはplayerID】【未使用】
                    if (playerID > 0) { KS_flagPlayer = true; }
                    d = ((d >> KS_PlayerIndexBitmaskWidth) & 0x1FFF);  //【realDepth】0mm～8191mm
                    if (d < KS_Lmin) { d = KS_Lmin; } //【近すぎる】
                    if (d > KS_Lmax) { d = KS_Lmax; } //【遠すぎる】
                    d = (int)((float)(d - KS_Lmin) * a) * 4;
                    KS_DepthImageDataARGB[j] = ColorBar.NTL_ColorElement[d]; //【青】B
                    KS_DepthImageDataARGB[j + 1] = ColorBar.NTL_ColorElement[d + 1]; //【緑】G
                    KS_DepthImageDataARGB[j + 2] = ColorBar.NTL_ColorElement[d + 2]; //【赤】R
                    KS_DepthImageDataARGB[j + 3] = ColorBar.NTL_ColorElement[d + 3]; //【α】
                    j = j + 4;
                }
            }
            else //【player検出に成功した場合にはPlayerのいる場所だけ表示して、他の部分は透明にする】
            {
                j = 0; f = 0;
                for (i = 0; i < KS_DepthData.Length; i++)
                {
                    d = KS_DepthData[i];                  //【距離情報】【上位13bitがDepth値】
                    playerID = d & KS_PlayerIndexBitmask; //【下位3bitはplayerID】【未使用】
                    if (playerID > 0)
                    {
                        f = 1;
                        d = ((d >> KS_PlayerIndexBitmaskWidth) & 0x1FFF);  //【realDepth】0mm～8191mm
                        if (d < KS_Lmin) { d = KS_Lmin; } //【近すぎる】
                        if (d > KS_Lmax) { d = KS_Lmax; } //【遠すぎる】
                        d = (int)((float)(d - KS_Lmin) * a) * 4;
                        KS_DepthImageDataARGB[j] = ColorBar.NTL_ColorElement[d];         //【青】B
                        KS_DepthImageDataARGB[j + 1] = ColorBar.NTL_ColorElement[d + 1]; //【緑】G
                        KS_DepthImageDataARGB[j + 2] = ColorBar.NTL_ColorElement[d + 2]; //【赤】R
                        KS_DepthImageDataARGB[j + 3] = ColorBar.NTL_ColorElement[d + 3]; //【α】
                    }
                    else
                    {   //【背景画像を透過させるために透明にする】
                        KS_DepthImageDataARGB[j] = 0;     //【青】B
                        KS_DepthImageDataARGB[j + 1] = 0; //【緑】G
                        KS_DepthImageDataARGB[j + 2] = 0; //【赤】R
                        KS_DepthImageDataARGB[j + 3] = 1; //【α】
                    }
                    j = j + 4;
                }
                if (f == 0) { KS_flagPlayer = false; } //【Playerがいなくなったようなのでモードを通常のデプスイメージ表示に戻す】
            }
        }

        private void radioButton3_CheckedChanged(object sender, EventArgs e)
        {
            pictureBox1.BackgroundImage = System.Drawing.Image.FromFile(BackGroundFileName[0]);
        }

        private void radioButton4_CheckedChanged(object sender, EventArgs e)
        {
            pictureBox1.BackgroundImage = System.Drawing.Image.FromFile(BackGroundFileName[1]);
        }

    }
}
